# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

'''
    A module containing the Module class declaration

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 01-06-2023 11:03:46
    `memberOf`: __init__
    `version`: 1.0
    `method_name`: Module
    * @TODO []: documentation for Module
'''


import datetime
from dataclasses import dataclass
import ast
from string import Template
from typing import Iterable

import colemen_utils as c

import silent.EntityBase as _eb

# import silent.Import.ImportStatement as _imp
from silent.Import import ImportStatement as _imp
from silent.DocBlock import ModuleDocBlock as _doc

import silent.Class as _class


import se_config as config
log = c.con.log
import silent.Method.Method as _method


@dataclass
class Module(_eb.EntityBase):


    _methods:Iterable[config._method_type] = None
    '''A list of methods contained in the module'''

    _classes:Iterable[config._py_class_type] = None
    '''A dictionary of classes contained in this module'''

    _imports:Iterable[config._py_import_type] = None
    '''A list of import statement instances.'''

    methods:Iterable[config._method_type] = None

    def __init__(self,main:config._main_type,package:config._package_type,**kwargs) -> None:
        '''
            Represents a python module.

            ----------

            Arguments
            -------------------------

            `main` {Main}
                A reference to the project instance.

            `package` {Package}
                A reference to the package this module belongs to.


            Keyword Arguments
            -------------------------

            `name` {str}
                The name of this module.

            `description` {str}
                The documentation description of this module


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-26-2022 08:35:16
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: Table
            * @xxx [12-26-2022 08:36:08]: documentation for Table
        '''
        kwargs['main'] = main
        kwargs['package'] = package
        super().__init__(**kwargs)



        self._classes = {}
        self._methods = []
        self._properties = {}
        self._imports = []
        self.Doc = _doc.ModuleDocBlock(self.main,self)




    @property
    def summary(self):
        '''
            Get the summary property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-06-2022 12:10:00
            `@memberOf`: __init__
            `@property`: summary
        '''
        value = {
            "name":self.name.name,
            "description":self.description,
            "imports":[],
            "classes":[],
            "methods":[],
            # "schema":self.table.database.database,
        }
        # value['imports'] = []
        for imp in self._imports:
            value['imports'].append(imp.summary)
        for cl in self.classes:
            value['classes'].append(cl.summary)
        for method in self.methods:
            value['methods'].append(method.summary)
        return value

    def save(self):
        c.file.write(self.file_path,self.result)

    # @property
    # def import_path(self):
    #     '''
    #         Get the import_path property's value

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-06-2023 12:31:41
    #         `@memberOf`: __init__
    #         `@property`: import_path
    #     '''
    #     value = self.relative_path.replace("/",".")
        
    #     return value

    # ---------------------------------------------------------------------------- #
    #                                    IMPORTS                                   #
    # ---------------------------------------------------------------------------- #


    def add_import(self,import_path=None,subjects=None,alias=None,is_standard_library=False,is_third_party=False)->config._py_import_type:
        '''
            Add an import statement to this module.

            ----------

            Arguments
            -------------------------

            [`import_path`=None] {str}
                The import path where the imported value is located.

            [`subjects`=None] {list,str}
                The subject or list of subjects to import

            [`alias`=None] {str}
                The alias to use for this import

            [`is_standard_library`=False] {bool}
                True if this import is from the python standard library.

            [`is_third_party`=False] {bool}
                True if this is importing from a third party library.



            Return {Import}
            ----------------------
            The new import statement instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-04-2023 10:19:47
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_import
            * @TODO []: documentation for add_import
        '''
        # print(f"add import: {subjects}")
        i = _imp(
            main=self.main,
            package=self.package,
            module=self,
            import_path=import_path,
            alias=alias,
            is_standard_library=is_standard_library,
            is_third_party=is_third_party,
        )
        i.add_subject(subjects)
        self._imports.append(i)
        return i

    @property
    def standard_library_imports(self)->Iterable[config._py_import_type]:
        '''
            Get the standard_library_imports property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 10:26:05
            `@memberOf`: __init__
            `@property`: standard_library_imports
        '''
        std_imports = []
        for imp in self._imports:
            print(f"import:{imp.subjects}")
            if imp.is_standard_library:
                std_imports.append(imp)
        return std_imports

    @property
    def third_party_imports(self)->Iterable[config._py_import_type]:
        '''
            Get a list of third party import statement instances associated to this module.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 10:26:05
            `@memberOf`: __init__
            `@property`: thrd_party_imports
        '''
        std_imports = []
        for imp in self._imports:
            if imp.is_third_party:
                std_imports.append(imp)
        return std_imports

    @property
    def local_imports(self)->Iterable[config._py_import_type]:
        '''
            Get the local_imports property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 10:26:05
            `@memberOf`: __init__
            `@property`: local_imports
        '''
        std_imports = []
        for imp in self._imports:
            if imp.is_standard_library is False and imp.is_third_party is False:
                std_imports.append(imp)
        return std_imports

    @property
    def _import_statements(self):
        '''
            Get the _import_statements property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 10:21:03
            `@memberOf`: __init__
            `@property`: _import_statements
        '''
        std_imports = '\n'.join(list([x.result for x in self.standard_library_imports]))
        thrd_imports = '\n'.join(list([x.result for x in self.third_party_imports]))
        local_imports = '\n'.join(list([x.result for x in self.local_imports]))

        value = f"{std_imports}\n{thrd_imports}\n{local_imports}"
        return value


    # ---------------------------------------------------------------------------- #
    #                                    CLASSES                                   #
    # ---------------------------------------------------------------------------- #


    def add_class(self,name:str,description:str=None,is_dataclass:bool=False)->config._py_class_type:
        '''
            Create a class instance and associate it to this module.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the class

            `description` {str}
                The documentation description of the class

            [`is_dataclass`=False] {bool}
                True if the class should be a dataclass.


            Return {Class}
            ----------------------
            The class instance

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-05-2023 11:30:17
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_class
            * @xxx [01-05-2023 11:36:41]: documentation for add_class
        '''
        nclass = _class.Class(
            self.main,
            module=self,
            name=name,
            description=description,
            is_dataclass=is_dataclass
        )
        self._classes[name] = nclass
        return nclass

    @property
    def classes(self)->Iterable[config._py_class_type]:
        '''
            Get a list of class instances associated to this module

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-04-2023 10:49:39
            `@memberOf`: __init__
            `@property`: classes
        '''
        value:Iterable[config._py_class_type] = list(self._classes.values())
        return value

    # @property
    # def _class_declarations(self):
    #     '''
    #         Get the _classes property's value

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-04-2023 10:48:15
    #         `@memberOf`: __init__
    #         `@property`: _classes
    #     '''
    #     value = []
    #     for mclass in self.classes:
    #         value.append(mclass.result)
    #     value= '\n\n\n'.join(value)
    #     return value


    # ---------------------------------------------------------------------------- #
    #                                    METHODS                                   #
    # ---------------------------------------------------------------------------- #


    def add_method(self,name,**kwargs)->config._method_type:
        '''
            Add a method to this class.
            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the method.

            Keyword Arguments
            -------------------------
            [`description`=None] {str}
                The docblock description for this method.

            [`body`="pass"] {str}
                The body of the method.

            [`return_type`=None] {str}
                The type returned by this method

            [`return_description`=None] {str}
                A description of the return value

            Return {Method}
            ----------------------
            The new method instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-04-2023 13:28:59
            `memberOf`: Module
            `version`: 1.0
            `method_name`: add_method
            * @xxx [01-06-2023 10:33:45]: documentation for add_method
        '''
        # kwargs['name'] = name
        method = _method.Method(
            self.main,
            module=self,
            package=self.package,
            name=name,
            description=c.obj.get_kwarg(['description'],None,(str),**kwargs),
            body=c.obj.get_kwarg(['body'],None,(str),**kwargs),
            return_type=c.obj.get_kwarg(['return_type'],None,(str),**kwargs),
            return_description=c.obj.get_kwarg(['return_description'],None,(str),**kwargs),
            is_getter=False,
            is_setter=False,
            is_class_method=False,
        )

        self._methods.append(method)
        return method


    @property
    def methods(self)->Iterable[config._method_type]:
        '''
            Get a list of methods that belong to this module.

            This does NOT include class methods only globally available methods.

            `default`:[]


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 11:24:35
            `@memberOf`: __init__
            `@property`: methods
        '''
        return self._methods


    @property
    def ast(self)->ast.Module:
        '''
            Get this modules ast object.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 11:03:02
            `@memberOf`: __init__
            `@property`: ast
        '''
        value = ast.Module(body=[],type_ignores=[])

        for imp in self._imports:
            value.body.append(imp.ast)

        # TODO []: apply global variables to body
        # xxx [01-06-2023 11:29:06]: apply classes to body
        for cl in self.classes:
            value.body.append(cl.ast)

        # xxx [01-06-2023 11:29:03]: apply functions to body
        for method in self.methods:
            value.body.append(method.declaration_ast)


        # @Mstep [] apply the module docblock to the body.
        value.body.insert(0,ast.Expr(value=ast.Constant(value=self.Doc.result)))
        value = ast.fix_missing_locations(value)
        return value

    @property
    def result(self):
        '''
            Get the result property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 11:49:00
            `@memberOf`: __init__
            `@property`: result
        '''
        value = ast.unparse(self.ast)
        return value

    # @property
    # def result(self):
    #     '''
    #         Get the result property's value

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 01-04-2023 09:34:33
    #         `@memberOf`: __init__
    #         `@property`: result
    #     '''
    #     module_name = self.name
    #     if module_name is None:
    #         if len(self.classes) > 0:
    #             clas:config._py_class_type = self.classes[0]
    #             module_name = clas.name

    #     description = "A module containing the stuff." if self.description is None else self.description



    #     s = Template(config.get_template("module_template"))
    #     value = s.substitute(
    #         classes=self._class_declarations,
    #         imports=self._import_statements,
    #         description=description,
    #         module_name=module_name,
    #         timestamp=datetime.datetime.today().strftime("%m-%d-%Y %H:%M:%S"),
    #     )
    #     # value = re.sub(r"^\s*$","\n",value)
    #     # value = re.sub(r"[\n]{3,}","\n\n",value)
    #     # value = re.sub(r"^[\t\s]\n","",value)

    #     return value






# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
# pylint: disable=import-outside-toplevel




import datetime
from dataclasses import dataclass
import re
from string import Template
from typing import Iterable, Union

import colemen_utils as c


# from silent.Import import ImportStatement as _imp
import silent.Import as _imp
# import silent.Module as _module


import se_config as config
log = c.con.log


@dataclass
class EntityBase:

    main:config._main_type = None
    '''A reference to the project instance.'''

    package:config._package_type = None
    '''The package that this entity belongs to.'''

    module:config._py_module_type = None
    '''The module that this entity belongs to.'''

    pyclass:config._py_class_type = None
    '''The class that this entity belongs to.'''

    method:config._method_type = None
    '''The class that this entity belongs to.'''

    _name:c.string.Name = None
    '''The name of this entity'''

    description:str = None
    '''The documentation description of this entity.'''

    _file_name:str = None
    '''The name of the file this entity resides in.'''
    _file_path:str = None
    '''The path to where this entity is saved.'''

    _tags:Iterable[str] = None



    def __init__(self,**kwargs):
        '''
            The base class used for all instances

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-26-2022 08:35:16
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: Table
            * @xxx [12-26-2022 08:36:08]: documentation for Table
        '''

        self._tags = []

        populate_from_dict(kwargs,self)
        if self.main is not None:
            self.main.register(self)


        
        from silent.Method import Method
        if isinstance(self,Method.Method):
            if self.pyclass is not None:
                self.module = self.pyclass.module
            if self.module is not None:
                self.package = self.module.package






    # @property
    # def summary(self):
    #     '''
    #         Get the summary property's value

    #         `default`:None


    #         Meta
    #         ----------
    #         `@author`: Colemen Atwood
    #         `@created`: 12-06-2022 12:10:00
    #         `@memberOf`: __init__
    #         `@property`: summary
    #     '''
    #     value = {
    #         # "schema":self.table.database.database,
    #         # "name":self.name.name,
    #     }

    #     return value

    @property
    def name(self)->c.string.Name:
        '''
            Get the name value.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 14:04:26
            `@memberOf`: PostArg
            `@property`: name
        '''
        value = self._name
        return value

    @name.setter
    def name(self,value:str):
        '''
            Set the name value.

            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 14:04:26
            `@memberOf`: PostArg
            `@property`: name
        '''
        if isinstance(value,(str)):
            self._name = c.string.Name(value)
        if isinstance(value,c.string.Name):
            self._name = value

    def has_tag(self,tag:Union[str,list],match_all=False):
        '''
            Check if this entity has matching tags.

            ----------

            Arguments
            -------------------------
            `tag` {str,list}
                A tag or list of tags to search for.

            [`match_all`=False] {bool}
                If True, all tags provided must be found.


            Return {bool}
            ----------------------
            True if the tag(s) are found.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-05-2023 09:29:33
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: has_tag
            * @xxx [01-05-2023 09:30:55]: documentation for has_tag
        '''
        tags = c.arr.force_list(tag)
        tags.sort()
        tags.sort(key=len, reverse=False)

        matched = []
        for tag in tags:
            if tag in self._tags:
                if match_all is False:
                    return True
                else:
                    matched.append(tag)
        diff = c.arr.find_list_diff(tags,matched)
        if len(diff) == 0:
            return True
        return False

    def add_tag(self,tag:Union[str,list],remove:Union[str,list]=None):
        '''
            Add a tag to this entity.
            ----------

            Arguments
            -------------------------
            `tag` {str,list}
                A tag or list of tags to add.
                This can also be a comma delimited string of tags

            `remove` {str,list}
                A tag or list of tags to remove.
                This can also be a comma delimited string of tags

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-05-2023 09:31:06
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_tag
            * @xxx [01-05-2023 09:31:39]: documentation for add_tag
        '''
        if isinstance(tag,(str)):
            tag = tag.split(",")

        tags = c.arr.force_list(tag)
        for tag in tags:
            if isinstance(tag,(str)) is False:
                raise Exception("Tags must be of type string.")
            self._tags.append(tag)
            self._tags = c.arr.remove_duplicates(self._tags)
            self._tags.sort()
            self._tags.sort(key=len, reverse=False)
        if remove is not None:
            self.delete_tag(remove)

    def delete_tag(self,tag:Union[str,list]):
        '''
            Delete a tag from this entity.
            ----------

            Arguments
            -------------------------
            `tag` {str,list}
                A tag or list of tags to delete.
                This can also be a comma delimited string of tags


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-05-2023 09:31:06
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: delete_tag
            * @xxx [01-05-2023 09:31:39]: documentation for delete_tag
        '''
        if isinstance(tag,(str)):
            tag = tag.split(",")

        rtags = c.arr.force_list(tag)
        output = []
        for tag in self._tags:
            if tag not in rtags:
                output.append(tag)
        # for tag in tags:
            # if tag not in self._tags:
                # output.append(tag)
        self._tags = output


    @property
    def file_name(self):
        '''
            Get the file_name property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 10:17:45
            `@memberOf`: __init__
            `@property`: file_name
        '''
        if self.name is not None:
            self._file_name = f"{self.name.name}.py"

        if self._file_name is not None:
            return self._file_name

        from silent.Method import Method
        if isinstance(self,Method.Method):
            return self.module.file_name

        from silent.Module import Module
        if isinstance(self,Module):
            if len(self.classes) > 0:
                value = f"{self.classes[0].name.name}.py"





        self._file_name = value
        return value


    @property
    def dir_path(self):
        '''
            Get the dir_path property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 10:20:19
            `@memberOf`: __init__
            `@property`: dir_path
        '''
        from silent.Package import Package
        if isinstance(self,Package):
            return self._file_path

        from silent.Module import Module
        if isinstance(self,Module):
            pkg = self.package
            return pkg.file_path

        from silent.Method import Method
        if isinstance(self,Method.Method):
            return self.module.file_path


    @property
    def file_path(self):
        '''
            Get the file_path property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 10:14:44
            `@memberOf`: __init__
            `@property`: file_path
        '''
        if self._file_path is not None:
            return self._file_path

        from silent.Package import Package
        if isinstance(self,Package):
            self._file_path = f"{self.dir_path}/__init__.py"

        from silent.Module import Module
        if isinstance(self,Module):
            self._file_path = f"{self.package.file_path}/{self.file_name}"

        from silent.Method import Method
        if isinstance(self,Method.Method):
            self.package = self.module.package
            self._file_path = f"{self.package.file_path}/{self.module.name.name}"

        from silent.Class import Class
        if isinstance(self,Class):
            self.package = self.module.package
            self._file_path = f"{self.package.file_path}/{self.module.name.name}"




        return self._file_path

    @file_path.setter
    def file_path(self,value):
        '''
            Set the file_path property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 09:57:27
            `@memberOf`: __init__
            `@property`: file_path
        '''
        self._file_path = value

    @property
    def relative_path(self):
        '''
            Get the relative_path property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 12:32:33
            `@memberOf`: __init__
            `@property`: relative_path
        '''

        value = self.file_path.replace(self.main.root_path,"")

        return value

    @property
    def import_path(self)->str:
        '''
            Get the import_path property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 12:44:16
            `@memberOf`: __init__
            `@property`: import_path
        '''
        value = None
        from silent.Module import Module
        if isinstance(self,Module):
            value = path_to_dot_path(self.relative_path)
        from silent.Package import Package
        if isinstance(self,Package):
            value = path_to_dot_path(self.relative_path)
        from silent.Method import Method
        if isinstance(self,Method.Method):
            value = path_to_dot_path(self.relative_path)
        return value

    @property
    def import_statement(self)->str:
        '''
            Get this entity's import statement as a string.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 13:22:04
            `@memberOf`: __init__
            `@property`: import_statement
        '''
        imp = _imp.ImportStatement()
        from silent.Package import Package
        if isinstance(self,Package):
            imp.add_subject(self.import_path)
            imp.alias = self.name.name
        else:
            imp.import_path=self.import_path
            imp.add_subject(self.name.name)
        value = imp.result
        return value







def populate_from_dict(data:dict,instance:EntityBase):
    for k,v in data.items():
        if hasattr(instance,k):
            if k == "name":
                if isinstance(v,(str)):
                    setattr(instance,'name',c.string.Name(v))
                if isinstance(v,c.string.Name):
                    setattr(instance,'name',v)
                continue
            setattr(instance,k,v)

def path_to_dot_path(relative_path:str):
    value = relative_path
    value = c.string.file_path(value,url=True)
    value = value.replace("/",".")
    value = c.string.strip(value,["."," "])
    value = c.string.strip(value,[".py"],"right")
    return value


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import




import datetime
from dataclasses import dataclass
# import re
from string import Template
from typing import Iterable

import colemen_utils as c


from silent.Import import ImportStatement as _imp
import silent.Module as _module
import silent.DocBlock.PackageDocBlock as _doc
import silent.EntityBase as _eb

import se_config as config
log = c.con.log


@dataclass
class Package(_eb.EntityBase):


    # file_path:str = None
    # '''The file path to this package.'''



    def __init__(self,main:config._main_type,**kwargs) -> None:
        '''
            Represents a python package.

            ----------

            Arguments
            -------------------------
            `main` {Main}
                A reference to the project instance.

            Keyword Arguments
            -------------------------
            `name` {str}
                The name of the module.

            `description` {str}
                The documentation description for the module.

            `file_path` {str}
                The file path to where this package will be saved.


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-26-2022 08:35:16
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: Table
            * @xxx [12-26-2022 08:36:08]: documentation for Table
        '''
        kwargs['main'] = main
        super().__init__(**kwargs)


        self._modules = {}
        self._imports:Iterable[config._py_import_type] = []


    @property
    def summary(self):
        '''
            Get the summary property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-06-2022 12:10:00
            `@memberOf`: __init__
            `@property`: summary
        '''

        value = {
            "name":self.name.name,
            "file_path":self.file_path,
            "tags":self._tags,
            "modules":{}
        }
        for mod in self.modules:
            _=mod.ast
            value['modules'][mod.name.name] = mod.summary


        return value


    def save(self):
        '''
            Save this package to the output directory.

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-06-2023 12:30:07
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: save
            * @xxx [01-06-2023 12:30:34]: documentation for save
        '''
        if c.dirs.exists(self.dir_path) is False:
            c.dirs.create(self.dir_path)

        mod_imports = []
        for mod in self.modules:
            mod_imports.append({"import_path":mod.import_path,"name":mod.name.name})
            mod.save()

        # print(f"mod_imports: {self.imports}")
        print(self.import_path)
        # c.file.write(self.file_path,self.init_result)
        c.file.write(f"{self.dir_path}/__init__.py",self.init_result)

    # ---------------------------------------------------------------------------- #
    #                                    MODULES                                   #
    # ---------------------------------------------------------------------------- #

    def add_module(self,name:str,description:str=None)->config._py_module_type:
        '''
            Create a module and associate it to this package.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the module.
            `description` {str}
                The documentation description for the module.


            Return {Module}
            ----------------------
            The new module instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-05-2023 10:11:02
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: add_module
            * @xxx [01-05-2023 10:12:04]: documentation for add_module
        '''
        mod = _module.Module(
            self.main,
            self,
            name=name,
            description=description,
        )
        self.add_module_import(mod)
        self._modules[name] = mod

        return mod

    @property
    def modules(self)->Iterable[config._py_module_type]:
        '''
            Get a list of module instances associated to this package.

            `default`:[]


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-05-2023 10:08:44
            `@memberOf`: __init__
            `@property`: modules
        '''
        value = list(self._modules.values())
        return value

    def get_module_by_name(self,name:str)->config._py_module_type:
        '''
            Retrieve a module by searching for its name.
            ----------

            Arguments
            -------------------------
            `name` {str}
                The module name to search for.


            Return {Module,None}
            ----------------------
            The module instance if it is found, None otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 01-05-2023 10:20:12
            `memberOf`: __init__
            `version`: 1.0
            `method_name`: get_module_by_name
            * @xxx [01-05-2023 10:21:02]: documentation for get_module_by_name
        '''
        for mod in self.modules:
            if mod.name.name == name:
                return mod
        return None


    def add_module_import(self,module:config._py_module_type):
        imp = _imp()
        if len(module.classes) == 0:
            imp = _imp(import_path=module.import_path)
            imp.add_subject("*")
        else:
            imp.add_subject(module.import_path)
            imp.alias = module.name.name
        
        self._imports.append(imp)

    @property
    def imports(self)->Iterable[config._py_import_type]:
        '''
            Get a list of import instances associated to this package.

            `default`:[]


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 12:50:31
            `@memberOf`: __init__
            `@property`: imports
        '''
        value = self._imports
        return value

    @property
    def init_result(self):
        '''
            Get the init_result property's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 01-06-2023 12:55:38
            `@memberOf`: __init__
            `@property`: init_result
        '''
        doc = _doc.PackageDocBlock(self.main,self)
        value = [doc.result,"\n\n"]
        for imp in self._imports:
            value.append(imp.result)
        value = '\n'.join(value)
        # value = self.init_result
        return value

# def populate_from_dict(data:dict,instance:Package):
#     for k,v in data.items():
#         if hasattr(instance,k):
#             setattr(instance,k,v)

